#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mach.h>
#include <dynld.h>

static ulong
_get2(uchar *b)
{
	return (b[0] << 8) | b[1];
}

static ulong
_get4(uchar *b)
{
	return (b[0] << 24) | (b[1] << 16) | (b[2] << 8) | b[3];
}

ulong
dyn_import_table(int fd, Fhdr *fh, Dynsym **table)
{
	int res = 0;
	uchar buf[4];
	vlong offset;
	ulong relsize, nimport, i;
	uchar *atab, *rp, *ep;
	Dynsym *tab;

	/* If not dlm, don't parse further */
	if(fh->magic != dynmagic())
		return 0;

	offset = seek(fd, 0, 1);
	seek(fd, fh->lnpcoff+fh->lnpcsz, 1);

	if(read(fd, buf, 4) != 4)
		goto end;
	relsize = _get4(buf);
	if (relsize < 4)
		goto end;

	if(read(fd, buf, 4) != 4)
		goto end;
	nimport = _get4(buf);
	if(nimport == 0)
		goto end;

	atab = malloc(nimport * sizeof(Dynsym) + relsize - 4);
	if(atab == nil)
		goto end;

	rp = atab + nimport * sizeof(Dynsym);
	ep = rp + relsize;
	tab = (Dynsym *)atab;

	if(read(fd, rp, relsize - 4) != relsize - 4)
		goto fail;

	for(i = 0; i < nimport; i++){
		if(rp +5 > ep)
			goto fail;
		tab[i].sig = _get4(rp);
		rp += 4;
		tab[i].name = (char *)rp;
		while(*rp++)
			if(rp >= ep)
				goto fail;
	}

	/* Trim off any relocation data we don't care about */
	*table = realloc(atab, rp - atab);
	res = nimport;

	goto end;

fail:
	free(atab);

end:
	seek(fd, offset, 0);
	return res;
}

ulong
dyn_reloc_table(int fd, Fhdr *fh, Dynrel **table)
{
	ulong res = 0;
	Dynrel *tab;
	uchar *edata, *reldata, *rp, *ep;
	ulong nimport, nreloc, i, p, a;
	ulong edsize, tsize, relsize;
	vlong offset;
	uchar buf[4];
	int m, n;

	/* If not dlm, don't parse further */
	if(fh->magic != dynmagic())
		return 0;

	/* Load dlm data */
	offset = seek(fd, 0, 1);
	seek(fd, fh->lnpcoff+fh->lnpcsz, 1);

	if(read(fd, buf, 4) != 4)
		goto end;

	relsize = _get4((uchar *)buf);
	if (relsize < 4)
		goto end;

	reldata = malloc(relsize);

	if(read(fd, reldata, relsize) != relsize)
		goto fail_reldata;

	nimport = _get4(reldata);
	ep = reldata + relsize;
	rp = reldata + 4;

	/* Find the relocation table */
	for(i = 0; i < nimport; i++){
		if(rp +5 > ep)
			goto fail_reldata;
		rp += 4;
		while(*rp++)
			if(rp >= ep)
				goto fail_reldata;
	}

	if(rp+4 > ep)
		goto fail_reldata;

	nreloc = _get4(rp);
	rp += 4;

	/* Read the file data, we're going to need this */
	seek(fd, offset + fh->txtoff, 0);

	edsize = fh->txtsz + fh->datsz;
	tsize = edsize + fh->bsssz;
	edata = mallocz(tsize, 1);
	if(edata == nil)
		goto end;

	if(read(fd, edata, edsize) != edsize)
		goto fail_edata;

	tab = malloc(nreloc * sizeof(Dynrel));

	/* Loop over the relocation table, filing in table entries */
	for(i = 0, a = 0; i < nreloc; i++){
		if(rp >= ep)
			goto fail;
		m = *rp++;
		n = m >> 6;
		if(rp + (1<<n) > ep)
			goto fail;
		switch(n) {
			case 0:
				p = *rp++;
				break;
			case 1:
				p = _get2(rp);
				rp += 2;
				break;
			case 2:
				p = _get4(rp);
				rp += 4;
				break;
			default:
				goto fail;
		}
		a += p;
		tab[i].offset = a;
		tab[i].mode = m & 0xf;
		if(crack_dynrel(edata, &tab[i]))
			goto fail;
	}

	*table = tab;
	res = nreloc;

	goto end;

fail:
	free(tab);
fail_edata:
	free(edata);
fail_reldata:
	free(reldata);
end:
	seek(fd, offset, 0);
	return res;

}